
#include "memcpy.c"
//#define my_memcpy memcpy6
#define my_memcpy memcpy
//#define my_memcpy memcpy_sse
//#define my_memcpy memcpy4


static double snapshot_time=0.0;

// Called (once) from gc_thread
static void calculate_snapshot_time(void){
  void *mem=my_calloc(1,heap_size);

  set_realtime(SCHED_FIFO,99);
  usleep(1);

  double verystart=get_ms();

  for(;;){
    double start=get_ms();
    my_memcpy(snapshot_mem,mem,heap_size);
    double time=get_ms()-start;
    if(snapshot_time==0.0 || time<snapshot_time)
      snapshot_time=time;
    if(start-verystart > 2000)
      break;
  }

  fprintf(stderr,"snapshot time: %fms (%f GB/s)\n",
          (float)snapshot_time,
          (float) (heap_size / snapshot_time) * (1000.0/(1024.0*1024.0*1024.0))
          );

  set_realtime(SCHED_OTHER,0);
  usleep(1);

  free(mem);
}


/*
take_heap_snapshot(heap):
  \new{if at_least_one_second_since_last_full_copy(current_heap):}
     \new{size = heap_size}
  \new{else:}
     \new{size = heap.mheap.freemem - heap.start}

  for i = 0 \new{size}:
      snapshot_mem[i] = heap.start[i]
*/


#if PARALLEL_SNAPSHOT
semaphore_t *heap_semaphore;
semaphore_t *heap_semaphore_finished;
void *par_from;
void *par_to;
size_t par_size;

static void par_snapshot_thread(void){
  heap_semaphore=semaphore_create(0);
  heap_semaphore_finished=semaphore_create(0);
  set_snapshot_thread_prio(1);
  while(true){
    semaphore_wait(heap_semaphore);
    my_memcpy(par_to,par_from,par_size);
    //my_memcpy(par_to,par_from,512*1024);
    semaphore_signal(heap_semaphore_finished);
  }
}
#endif

// Actually, somewhere between half and full snapshot.
static void take_full_snapshot(tar_heap_t *heap){
  char *dest=snapshot_mem;
  char *src=(char*)GET_RAW_MEM(heap->mheap);
  double start=get_ms();
  int last_size=0;
  int so_far=GET_CONT_HEAP_SIZE(heap->mheap,heap_size);

  if(half_heap_size > so_far)
    so_far=half_heap_size;

  my_memcpy(dest,src,so_far);

  for(;;){

    double time=get_ms() - start;
    if(false
       || time >= snapshot_time
       || so_far == heap_size
       )
      {
	//fprintf(stderr,"size: %d\n",so_far);
	return;
      }

    last_size = so_far;
    so_far += SNAPSHOT_COPY_SIZE;
    if(so_far > heap_size)
      so_far = heap_size;

    int size = so_far - last_size;
    my_memcpy(dest+last_size,src+last_size,size);
  }

}

// Called from gc_thread
#if PARALLEL_SNAPSHOT
#warning PARALLEL_SNAPSHOT has unpredictable execution time
static void take_heap_snapshot(tar_heap_t *heap){
  size_t size =
    heap->take_full_snapshot 
    ? heap_size 
    : GET_CONT_HEAP_SIZE(heap->mheap,heap_size);


  if(heap->take_full_snapshot){
    int half=size/2;
    char *from=GET_RAW_MEM(heap->mheap);
    par_to=snapshot_mem+half;
    par_from=from+half;
    par_size=size-half;
    semaphore_signal(heap_semaphore);
    my_memcpy(snapshot_mem,from,half);
    //my_memcpy(snapshot_mem,from,512*1024);
    semaphore_wait(heap_semaphore_finished);
  }else
    my_memcpy(snapshot_mem,GET_RAW_MEM(heap->mheap),size);
}
#else
static void take_heap_snapshot(tar_heap_t *heap){
  if(heap->take_full_snapshot)
    take_full_snapshot(heap);
  else
    my_memcpy(snapshot_mem,GET_RAW_MEM(heap->mheap),GET_CONT_HEAP_SIZE(heap->mheap,heap_size));
}
#endif


